#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2011, 2013 All Rights Reserved
#*
# hostvolattach.sh [-l] [-m] [--ids <WWID>:<LUN>~<address>~<blocks>~<initiator>~<target> ...]
#   version: 2.5 [5/06/2013]
#   author:  csky
#
# !Script only supported on RHEL 6!
#
#    Description: Scan scsi bus HBA's for fresh set of volumes. Optionally look
#                 for a specific set. Dump out info for the set or all disks found.
#                 Optionally check first for MPIO stable paths (e.g. /dev/mapper/...)
#                 Call a fixing script if any of the specific set of volumes are
#                 not located.
#    Options:
#        -l    long (verbose)
#        -m    multi-path enabled
#       --ids  All arguments following this option are treated as a specific volume
#              reference, consisting of a WWID (DeviceID) and LUN combination
#              separated by a ':'. If specified, then only those devices are fully
#              processed. If the LUN is unknown, pass -1. Everything following the LUN
#              is also optional. These secondary properties are separated by a '~'.
#   Format: <WWID>:<LUN>~<address>~<blocks>~<initiator-list>~<target-list>
#   Example: 60050768028180ee380000000000500f:5~IBM_Storwize_v7000-IBM/v7000_IBM_Mid-Range_DS5300_Pool/myVolName~2048~0x2100001b329eca3c,0x2101001b329eca3c~0x201700a0b86e33ca
#              address: The external named path to the storage (not the host view).
#              blocks: Size of vol in number of 512 byte blocks
#              initiatior-list: The hba port WWPNs if applicable
#              target-list: The transport target controller WWPNs if applicable

cmdwait=15
scanwait=120
checkwait=500
vmcdir="/var/opt/ibm/vmc"
fixlogfile="${vmcdir}/outsancheck.trace"
alogfile="${vmcdir}/outattach.trace"
argids=N
long=false
fixmode="id-enabled"
err_codes=""

#subroutines
log_msg(){
	local msg=$1
	echo "[$$] $msg" >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "$msg"
	fi
}

log_err(){
	local msg=$1
	dstamp=$(date)
	echo "[$$]$dstamp: $msg" >> $alogfile
	# 1>&2 sends stdout to stderr
	echo "[$$]$dstamp: $msg" 1>&2
}

log_print(){
    local msg=$1
	echo "[$$] $msg" >> $alogfile
	echo "[$$] $msg"
}

get_scsiid(){
	log_msg "  scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1"
	scsiid=$(timeout -s SIGKILL $cmdwait scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1)
	local sret=$?
	if [ "$sret" = "0" ]
	then
		scsiid=${scsiid##[ ]*}
		return 0
	else
		scsiid="-"
		err_codes="$err_codes s${sret}\n"
		log_err "[scsi_id return for device=$1  **Command failed. [rc=$sret]"
		return $sret
	fi
}

dd_read(){
	log_msg "  dd if=$1 of=/dev/null bs=32 count=1 2>&1"
	chkread=$(timeout -s SIGKILL $cmdwait dd if=$1 of=/dev/null bs=32 count=1 2>&1)
	local sret=$?
	if [ "$sret" != "0" ]
	then
		err_codes="$err_codes d${sret}\n"
	fi
	log_msg "  dd reading $1 [rc=$sret]   output=$chkread"
	return $sret
}

# The "locking" done in this function is not to provide true synchronization.
# It just helps alleviate the load if many attach requests come in together so
# that the looping and processing through the volumes does not happen at once and
# bog down the system.
sancheck(){
	local sargs=$1
	local logto=$2
	local dstamp=$(date)
	local loop=0
	log_msg "Check for lck file. $sargs [$3]"
	if [ -f $vmcdir/lck.sancheck ]
	then
		log_msg "$dstamp ***sancheck lock file FOUND. Wait till it is gone or too old or starved out..."
		while [ -f $vmcdir/lck.sancheck ]
		do
			old=$(( (`date +%s` - `stat -L --format %Y $vmcdir/lck.sancheck`) > (30*60) ))
			log_msg "lck check: old=$old"
			if [ "$old" = "1" ]
			then
				log_msg "Removing old lock file and continuing..."
				rm -f $vmcdir/lck.sancheck
			fi
			contents=$(cat $vmcdir/lck.sancheck)
			sleep 1
			let loop=loop+1
			if [ $loop -gt $checkwait ]
			then
				log_msg "!Starved out waiting for sancheck lock! Loop=$loop"
				break
			fi
		done
		if [[ "$sargs" == *ids* ]]
		then
			log_msg "Specific vols passed to fix. Continue with sancheck..."
		else
			if [[ "$contents" == *ids* ]]
			then
				log_msg "Continue with general volume cleanup requested. Prior cleanup was for specific vols."
			else
				log_msg "Prior cleanup requested was general and since this is general, just return."
				return 0
			fi
		fi
		marker=n
		[ -f "$vmcdir/stale.sancheck" ] && marker=y
		if [ "$3" = "stale" -a "$marker" = "n" ]
		then
			log_msg "Stale marker no longer exists. Just return."
			return 0
		else
			log_msg "Stale marker still there. Continue with cleanup."
		fi
	fi

	echo $sargs >> $vmcdir/lck.sancheck
	local lines=$(cat $vmcdir/lck.sancheck | wc -l)
	log_msg "Touched lock file: $vmcdir/lck.sancheck  [lines=$lines]"
	dstamp=$(date)
	log_msg "  $dstamp: timeout -s SIGKILL 150 $vmcdir/sancheck.sh $sargs"
	timeout -s SIGKILL $checkwait $vmcdir/sancheck.sh $sargs >> $logto
	local sret=$?
	dstamp=$(date)
	echo "  $dstamp   *** sancheck numfixed/rc=$sret ***" >> $logto
	log_msg "  $dstamp   *** sancheck numfixed/rc=$sret ***"
	rm -f $vmcdir/lck.sancheck
	log_msg "Removed lock file: $vmcdir/lck.sancheck"
	if [ "$3" = "stale" ]
	then
		rm -f "$vmcdir/stale.sancheck"
		log_msg "Removed any $vmcdir/stale.sancheck"
	fi
	return $sret
}

hba_scan(){
	local dstamp=$(date)
	local hbas="$2"
	local itl="- - $1"
	# Do we want to synchronize the scans and/or prevent scans that happened within the last couple seconds?
	log_msg "  $dstamp: timeout -s SIGKILL $scanwait echo $itl | tee $hbas > /dev/null"
	timeout -s SIGKILL $scanwait echo $itl | tee $hbas > /dev/null
	local sret=$?
	dstamp=$(date)
	log_msg "  $dstamp *** hba scsi bus scan   rc=$sret ***"
	if [ "$sret" != 0 ]
	then
		err_codes="$err_codes d${sret}\n"
	else
		sleep 1
		log_msg "Issue udevadm settle as a synchronization point for udev symlink processing to catch-up..."
		udevadm settle
		local ret=$?
		dstamp=$(date)
		log_msg "  $dstamp *** udevadm settle   rc=$ret ***"
    fi
	return $sret
}

chk_ports_transp() {
	numOnline=$(cat /sys/class/fc_host/host*/port_state | grep -c Online)
	if [ "$numOnline" = "0" ]
	then
		log_err "None of the HBA ports have a status of Online."
		errat="Y"
	fi

	# Check for FC transport
	numTransports=$(ls -1 /sys/class/fc_transport/ | wc -l)
	if [ "$numTransports" = "0" ]
	then
		log_err "No entries in /sys/class/fc_transport. Check zoning."
		errat="Y"
	fi
	log_msg "Number of online HBAs=$numOnline, Number of FC transports=$numTransports"
}

get_vol_inventory(){
	dstamp=$(date)
	log_msg "General volume inventory collection. First identify known volumes. $dstamp"
	idlistfile=$(mktemp $vmcdir/idlist.XXXXXXXX)
	for fcdir in $(ls -1 /sys/class/fc_transport)
	do
		target=$(echo $fcdir | cut -d : -f 1)
		channel=$(echo $fcdir | cut -d : -f 2)
		controller=$(echo $fcdir | cut -d : -f 3)
		hostnum=${target#target}
		#echo "Dir: /sys/class/fc_transport/$fcdir/device"
		for file in $(ls -1 /sys/class/fc_transport/$fcdir/device | egrep "^$hostnum:$channel:$controller:")
		do
			dlink=/sys/class/fc_transport/$fcdir/device/$file
			if [ -d $dlink/block ]
			then
				blockdev=$(ls $dlink/block)
				#echo "$file: $blockdev"
				devblockpath=/dev/$blockdev
				get_scsiid $devblockpath
				ret=$?
				if [ ${#scsiid} -gt 2 -o "$ret" != "0" ]
				then
					volid=${scsiid:1}
					volid=${volid##[ ]*}
					if [ -n "$volid" ]
					then
						# Avoid duplicates so just take the first device with the wwid
						ogrep=$(grep $volid $idlistfile)
						rc=$?
						if [ "$rc" != "0" ]
						then
							mapline=$(awk -v ss="$volid" '$0 ~ ss {print $17,$13,$5}' $vmcdir/pvmap | sort -r | head -n 1)
							pvlun=$(echo $mapline | cut -d ' ' -f 3)
							if [ -n "$pvlun" ]
							then
								pvdevname=$(echo $mapline | cut -d ' ' -f 2)
								if [[ "$pvdevname" != /dev/* ]]; then
									# does not start with /dev/ so prior symbolic name is preferred
									devblockpath=$pvdevname
								fi
							else
								pvlun=-1
							fi
							echo "${volid}:$pvlun~$devblockpath" >> $idlistfile
						fi
					fi
				else
					log_msg "Skipping device $devblockpath because unsupported scsiid format or scsi_id err: id=$scsiid ret=$ret"
				fi
			fi
		done
	done
	volumes=$(cat $idlistfile)
	dstamp=$(date)
	log_msg "Volume inventory set: *** $dstamp ***"
	log_msg "$volumes"
	rm -f $idlistfile
}

found_stable(){
	local pathpre=$1
	local cutfield=$2
	log_msg "Stable device: $stable"
	dev=$(echo $stable | cut -d '/' -f $cutfield)
	devpath=/dev/$dev
	stablename=$(echo $stable | cut -d ' ' -f 1)
	stablepath=$pathpre/$stablename
	log_msg "Stable path: $stablepath"
	volfound=Y
}

parse_vol_input(){
	# Checking for new I-T-L parms in preparation for them being passed
	local curvol=$1
	argwwid=${curvol%%:*}
	local v1a=${curvol##*:}
	arglun=${v1a%%~*}
	argblocks="-1"
	argname="[unknown-address]"
	arghba="-1"
	argtarget="-1"
	if [ "$arglun" = "$argwwid" ]
	then
		arglun="-1"
		return 0
	fi
	local v1b=${v1a#*~}
	if [ "$v1b" = "$v1a" ]
	then
		return 0
	fi
	argname=${v1b%%~*}
	local v1c=${v1b#*~}
	if [ "$v1c" = "$v1b" ]
	then
		return 0
	fi
	argblocks=${v1c%%~*}
	local v1d=${v1c#*~}
	if [ "$v1d" = "$v1c" ]
	then
		return 0
	fi
	arghba=${v1d%%~*}
	# arghba (scsi initiators) will likely be a list and so not used.
	local v1e=${v1d#*~}
	if [ "$v1e" = "$v1d" ]
	then
		return 0
	fi
	# argtarget (scsi target/controller) will likely be a list and so not used.
	argtarget=${v1d%%~*}
	return 0
}

cleanup_by_lun(){
	# only for MPIO
	local lret=1
	local rc=0
	for vol in $volumes
	do
		log_msg "======================================================"
		let numvols+=1
		# Split volume arguments in pieces
		parse_vol_input "$vol"
		log_msg "VOL cleanup check: wwid=$argwwid   blocks=$argblocks   Initiater/Target/Lun=$arghba/$argtarget/$arglun"
		if [ "$arglun" = "-1" ]
		then
			continue
		fi
		local first=$(multipath -l | awk -F: -v ss="[0-9]+[:][0-9]+[:][0-9]+[:]$arglun[ \t]+" '$0 ~ ss {print $4}' | head -n 1)
		log_msg "First line from multipath output matching x:x:x:$arglun --> $first"
		if [ -n "$first" ]
		then
			local context=$(multipath -l | grep -B 3 "$first")
			local foundwwid=${context%% dm-*}
			if [ -z "$foundwwid" ]
			then
				log_msg "No wwid found for dm-* in multipath context. Likely a timeing window. Continuing..."
				continue
			fi
			local founddev=$(echo $context | cut -d' ' -f 2)
			log_msg "LUN $arglun matches mpio device $founddev with wwid $foundwwid. Full context=$context"
			echo $foundwwid | grep -i $argwwid > /dev/null
			rc=$?
			if [ "$rc" != "0" ]
			then
				log_msg "wwids do not match, so do stale disk cleanup for LUN $arglun"
				lret=0
			else
				local checkout=$(dd if=/dev/$founddev of=/dev/null bs=32 count=1 2>&1)
				rc=$?
				if [ "$rc" != "0" ]
				then
					log_msg "Cannot read from /dev/$founddev (rc=$rc) so treating device as stale for LUN $arglun"
					lret=0
				fi
			fi
			if [ "$lret" = "0" ]
			then
				#First handle the legacy devices
				legdevs=""
				for legdev in $(multipath -l | awk -F: -v ss="[0-9]+[:][0-9]+[:][0-9]+[:]$arglun[ \t]+" '$0 ~ ss {print $4}' | awk '{print $2}')
				do
					echo "1" > /sys/block/$legdev/device/delete
					rc=$?
					log_msg "Ran: echo 1 > /sys/block/$legdev/device/delete  [rc=$rc]"
					if [ -z "$legdevs" ]
					then
						legdevs=$legdev
					else
						legdevs="$legdevs $legdev"
					fi
				done
				# At this point, the multipath device is probably gone, but we check here
				if [ -b /dev/$founddev ]
				then
					# flush MPIO dev - only works if not in use
					multipath -f /dev/$founddev >> $alogfile
					rc=$?
					log_msg "Ran: multipath -f /dev/$founddev  [rc=$rc]"
				fi
				# pause 3 seconds
				sleep 3
				# remap/scan devices
				hba_scan "$arglun" "$hbaHosts"
				rc=$?
				log_msg "hba_scan for LUN $arglun returned $rc."
				# At this point, the multipath device should be recreated. If not, run multipath by itelf.
				ls -1 /dev/mapper/ | grep -i $argwwid > /dev/null
				rc=$?
				if [ "$rc" = "0" ]
				then
					log_msg "MPIO symbolic link found for $argwwid."
				else
					multipath >> $alogfile
					rc=$?
					log_msg "'multipath' ran to map any new devices and the output was logged."
				fi
			else
				log_msg "A MPIO path was found for $argwwid and was readable, so nothing to do."
			fi
		else
			log_msg "No legacy scsi devices surfaced by multipath for LUN. Continue."
		fi
	done
	return $lret
}

find_specified_vols(){
	let numvols=0
	let numfound=0
	listnotfound=""
	dstamp=$(date)
	log_msg "find_specified_vols: $dstamp"
	if [ "$pvmapfile" != "" ]
	then
		rm -f $pvmapfile
	fi
	pvmapfile=$(mktemp $vmcdir/pvmap.XXXXXXXX)
	# Physical Volume lookup loop
	for vol in $volumes
	do
		log_msg "======================================================"
		log_msg "Volume: $vol"
		let numvols+=1
		volfound=N
		mpath=N
		health=OK
		# Split volume arguments in pieces
		parse_vol_input "$vol"
		if [ "$argname" = "-1" -o "$argname" = "[unknown-address]" ]
		then
			mapline=$(awk -v ss="$argwwid" '$0 ~ ss {print $17,$13,$5}' $vmcdir/pvmap | sort -r | head -n 1)
			mapname=$(echo $mapline | cut -d ' ' -f 2)
			if [ -n $mapname ]
			then
				argname="$mapname"
			else
				argname="/dev/unknown"
			fi
		fi
		log_msg "        wwid=$argwwid   blocks=$argblocks   Initiater/Target/Lun=$arghba/$argtarget/$arglun"
		# The SCSI initiator and target WWPNs are really only useful for probing when different
		# HBAs are zoned to different storage subsystems, which is not typical and this script does not
		# handle yet. For the LUN however, we will attempt specific fast path probing.
		if [ "$arglun" != "-1" -a "$argids" = "Y" ]
		then
			hba_scan "$arglun" "$hbaHosts"
			# Now issue multipath command to sync up multipaths - don't rely just on multipathd
			multipath >> $alogfile
			rc=$?
			log_msg "+++multipath cmd completed. rc=$rc"
		fi
		if [ "$targetedMpathd" = "Y" ]
		then
			#Special wait loop just for mpio /dev/mapper links
			#Should not need this loop now with udevadm settle fencing in place, but keeping in case we benefit on very loaded systems
			let xx=0
			while [ $xx -lt 11 ]
			do
				stable=$(ls -l /dev/mapper | awk -v ss="[^p]$argwwid[^p]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					found_stable "/dev/mapper" "2"
					log_msg "Exiting mpio wait loop for $dev."
					mpath=Y
					break
				fi
				let xx=$xx+2
				if [ $xx -lt 11 ]
				then
					sleep 2
					log_msg "Slept $xx seconds of 10. Looking for /dev/mapper link again."
				fi
			done
		fi
		# Search in priority order
		if [ "$mpatharg" = "Y" -a "$volfound" = "N" ]
		then
			# 1. for multipaths, search /dev/mapper
			stable=$(ls -l /dev/mapper | awk -v ss="[^p]$argwwid[^p]" '$0 ~ ss {print $9, $11}')
			if [ -n "$stable" ]
			then
				found_stable "/dev/mapper" "2"
				mpath=Y
			else
				# fallback to by-id, just in case user friendly names are turned on by mistake
				stable=$(ls -l /dev/disk/by-id | awk -v ss="dm-uuid-mpath-[0-9]$argwwid" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					found_stable "/dev/disk/by-id" "3"
					mpath=Y
				fi
			fi
		fi
		if [ "$volfound" = "N" ]
		then
			# wwn- prefix check (non-multipath)
			stable=$(ls -l /dev/disk/by-id | awk -v ss="wwn-0x$argwwid[^-]" '$0 ~ ss {print $9, $11}')
			if [ -n "$stable" ]
			then
				found_stable "/dev/disk/by-id" "3"
			else
				# fallback - allow any scsi type prefix that is a single digit, not just '3' ('scsi-.' below)
				stable=$(ls -l /dev/disk/by-id | awk -v ss="scsi-.$argwwid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					found_stable "/dev/disk/by-id" "3"
				fi
			fi
		fi
		if [ "$volfound" = "Y" ]
		then
			# Common 'disk' pieces
			blockdir=/sys/block/$dev
			if [ "$mpath" = "Y" ]
			then
				# For multipath output, assume the "active" path is 3 lines below the wwid entry. returns "5 sdaq 66"
				triple=$(multipath -l | grep -A 3 -i $argwwid | tail -n 1 | awk -F: '{print $4}')
				if [ -z "$triple" ]
				then
					log_msg "warning: $dev might have a link in /dev/mapper, but multipath does not report it. Perhaps it was attached at one time but is detached now. Flag volume as not found and continue search."
					log_msg "  **Volume not found: $argwwid  lun=$arglun  address=$argname"
					listnotfound="$listnotfound ${argwwid}:${arglun}~${argname}"
					volfound=N
					continue
				fi
				lun=$(echo $triple | cut -d' ' -f 1)
				legacyscsi=$(echo $triple | awk '{print $2}')
				# The scsidev (e.g. 1:0:2:3) is used to index into the sysfs device path
				dlink=/sys/block/$legacyscsi/device
				scsidev="*:*:*:$lun"
				hostwwpn="multi-wwpns-$numHosts"
			else
				# Non-multipath case. The Device dir is right underneath the $blockdir
				dlink=$blockdir/device
				scsidev=$(ls $dlink/scsi_device)
				scsi3dev=${scsidev%:*}
				lun=$(echo $scsidev | cut -d : -f 4)
				legacyscsi=$dev
				transportdir=/sys/class/fc_transport/target$scsi3dev
				if [ -d $transportdir ]
				then
					hostwwpn=$(cat $transportdir/port_name)
				else
					hostwwpn="NON-FC-TRANSPORT"
				fi
			fi
			log_msg "found lun=$lun, legacyScsi=$legacyscsi, deviceLink=$dlink"
			let numfound+=1
			blocks=$(cat $blockdir/size)
			if [ "$blocks" = "" ]
			then
				blocks="0"
			else
				if [ "$argblocks" != "-1" -a "$argblocks" != "$blocks" ]
				then
					log_msg "Warning: Number of blocks, $blocks, mismatch from what was passed in: $argblocks."
					log_msg "echo 1 > $dlink/rescan; sleep 1"
					echo "1" > $dlink/rescan
					sleep 1
					blocks=$(cat $blockdir/size)
					if [ "$argblocks" != "$blocks" ]
					then
						health="stale-size"
						log_msg "Rescan of the scsi device did not appear to help the issue. Blocks=$blocks from $blockdir/size"
					fi
				else
					if [ "$argblocks" != "-1" ]
					then
						log_msg "Num 512 byte blocks matches number expected: $argblocks"
					fi
				fi
			fi
			blocksize=$(cat $blockdir/queue/physical_block_size)
			if [ "$blocksize" = "" ]
			then
				blocksize="0"
			fi
			sizeGiB=$(echo "scale=3; ($blocks * $blocksize) / ( 1024 * 1024 * 1024 )" | bc)
			
			# Test device
			dd_read $devpath
			rc=$?
			if [ "$rc" != "0" ]
			then
				health="read-failure"
				log_msg "Warning: Could not read from device that was found $devpath"
			fi
			ronly=$(cat $blockdir/ro)
			access=$ronly
			if [ "$ronly" = "0" ]
			then
				#read/write
				access="3"
			else
				if [ "$ronly" != "1" ]
				then
					#unknown
					access="0"
				fi
			fi
			
			# Common 'device' pieces
			vendor=$(cat $dlink/vendor)
			vendor=${vendor%%[ ]*}
			model=$(cat $dlink/model)
			model=${model%%[ ]*}
			
			# send upper case wwid back
			deviceID=$(echo $argwwid | tr '[a-z]' '[A-Z]')
			# validation
			if [[ "$argname" == /dev/* ]]; then
				argname=$devpath
			fi
			if [ "$hostwwpn" = "" ]
			then
				hostwwpn="unknown"
			fi
			if [ "$vendor" = "" ]
			then
				vendor="unknown"
			fi
			if [ "$model" = "" ]
			then
				model="unknown"
			fi
			if [ -n "$devpath" -a -b $devpath ]
			then
				curdate=$(stat -c %y $devpath | awk '{print $1 "_" $2}')
				curdate=${curdate%%.*}
			else
				curdate="unknown"
			fi
			vsuse="---"
			if [ "$numvses" != "0" ]
			then
				filematch=$(grep $stablepath /etc/libvirt/qemu/*.xml | head -n 1 | cut -d : -f 1)
				if [ -n "$filematch" ]
				then
					vsuse=${filematch%%.xml}
					vsuse=${vsuse##*/}
				fi
			fi
			# wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health  device-date  using-vs
			volout=$(printf '%-33s %-70s %12s %11s %4s %4s %10s %1s %1s %18s %8s %8s %s %s  %s  %s  %s  %s' $deviceID $stablepath $devpath $scsidev $arglun $lun $blocks $mpath "$access" $hostwwpn $vendor $model $argname $blocksize $sizeGiB $health $curdate $vsuse)
			if [ "$hostwwpn" != "NON-FC-TRANSPORT" -o "$argids" = "Y" ]
			then
				# Only add this one if FC transport device
				echo "$volout" >> $pvmapfile
			fi
			log_msg "  ------------- Found volume $numfound -------------------"
			log_msg "$volout"
		else
			log_msg "  **Volume not found: $argwwid  lun=$arglun  address=$argname"
			listnotfound="$listnotfound ${argwwid}:${arglun}~${argname}"
			# keep going to get full list to fix
		fi
		# On to next volume to find
	done
}
# end volume search method
#=========================
# Mainline starts here.
#=========================
# This folder should exist since normally the scripts are contained in it.
if [ -d $vmcdir ]
then
	log_msg ""
else
	mkdir -p $vmcdir
	rc=$?
	if [ "$long" = "true" ]
	then
		echo "Created directory $vmcdir. rc=$rc"
	fi
fi

if [ -f $alogfile ]
then
	ccrc=$(find $alogfile -size +3000)
	if [ "$ccrc" = "$alogfile" ]
	then
		echo "Cleanup old log file (wrap to backup)."
		mv -f $alogfile "${alogfile}.1"
	fi
fi

# Process script arguments
if [ "$1" = "-l" -o "$2" = "-l" ]
then
	long=true
fi
mpatharg=N
mpathd=N
targetedMpathd=N
if [ "$1" = "-m" -o "$2" = "-m" ]
then
	mpatharg=Y
	service multipathd status | grep "is running" > /dev/null
	ret=$?
	if [ "$ret" = "0" ]
	then
		mpathd=Y
	fi
fi

if [ "$1" = "--ids" ]
then
	shift 1
	argids=Y
else
	if [ "$2" = "--ids" ]
	then
		shift 2
		argids=Y
	else
		if [ "$3" = "--ids" ]
		then
			shift 3
			argids=Y
		fi
	fi
fi


dstamp=$(date)
log_msg "$dstamp"
log_msg "long=$long   argids=$argids   mpatharg=$mpatharg   mpathd=$mpathd"
touch $vmcdir/pvmap

if [ "$argids" = "Y" ]
then
	volumes="$@"
	log_msg "Volumes= $volumes"
	if [ "$mpathd" = "Y" ]
	then
		#For scoping new code
		targetedMpathd=Y
	fi
fi

# Check HBA ports
numHosts=$(ls -1 /sys/class/fc_host/ | wc -l)
if [ "$numHosts" = "0" ]
then
	log_err "This endpoint does not appear to be attached to a Fibre Channel network. No entries in /sys/class/fc_host. Exiting."
	log_msg "exit 1"
	exit 1
fi
# compute FC HBAs
hbaHosts=$(ls /sys/class/fc_host/ | sed 's/host[0-9]*/\/sys\/class\/scsi_host\/&\/scan/g' | tr "\n" " ")
log_msg "Number of FC HBAs=$numHosts ($hbaHosts)"
# Do the "light" scan of scsi interconnect (the only semi-safe way) if general vol inv
if [ "$argids" = "N" ]
then
	hba_scan "-" "$hbaHosts"
	if [ "$mpathd" = "Y" -a "$mpatharg" = "Y" ]
	then
		# Now issue multipath command to sync up multipaths - don't rely just on multipathd
		multipath >> $alogfile
		rc=$?
		log_msg "+++multipath cmd completed for inv. rc=$rc"
	fi
fi
errat="N"
chk_ports_transp
if [  "$errat" = "Y" ]
then
	if [ "$argids" = "Y" ]
	then
		hba_scan "-" "$hbaHosts"
		log_msg "Check again - did FC ports/transports show up?"
		sleep 1
		errat="N"
		chk_ports_transp
	fi
	if [ "$errat" = "Y" ]
	then
		# error message already logged
		log_msg "exit 1"
		exit 1
	fi
fi

# Check for file existence to indicate whether a sancheck cleanup run should be done prior to scanning
if [ -f "$vmcdir/stale.sancheck" -a "$argids" = "N" ]
then
	bv=$(cat $vmcdir/stale.sancheck)
	log_msg "  Inventory pre SAN check & fix (staleVols=$bv)"
	sancheck "-fn" "$vmcdir/outinvsancheck.trace" "stale"
	rc=$?
	log_msg "  Inventory pre SAN check complete."
fi

if [ -f $fixlogfile ]
then
	ccrc=$(find $fixlogfile -size +3000)
	if [ "$ccrc" = "$fixlogfile" ]
	then
		log_msg "Cleanup old log file (wrap to backup)."
		mv -f $fixlogfile "${fixlogfile}.1"
	fi
fi

# Cleanup old temp map files older than a week
log_msg "Cleanup old temporary mapping files: "
find $vmcdir/pvmap.* -mtime +7 -exec rm -f {} \; 2> /dev/null
find $vmcdir/devmap.* -mtime +7 -exec rm -f {} \; 2> /dev/null
numvses=0
if [ -d /etc/libvirt/qemu ]
then
	numvses=$(find /etc/libvirt/qemu -maxdepth 1 -name '*.xml' | wc -l)
fi
dropThrough=N
let innertries=1
#=========================
# Outer loop - fix mode
#=========================
while [ "$fixmode" != "done" ]
do
	log_msg "Begin outer loop. Fixmode=$fixmode."
	if [ "$argids" = "N" ]
	then
		# Specific volumes not passed in, so process all current volumes
		get_vol_inventory
	fi

	let numtimes=0
	firsttime="true"
	# Timeout loop - waiting for devices to appear
	while [ $numtimes -lt $innertries ]
	do
		let numtimes+=1
		log_msg "Begin volume wait loop number $numtimes..."
		if [ "$firsttime" = "false" ]
		then
			log_msg "Pause 10..."
			sleep 10
		fi
		firsttime="false"
		find_specified_vols

		# End PV loop
		if [ "$numfound" = "$numvols" ]
		then
		    dstamp=`date`
			log_msg " *** $dstamp *** Number found matches the number sought. Setting fixmode=done"
			fixmode="done"
			break
		fi
		if [ "$dropThrough" = "Y" ]
		then
			log_msg "Drop out of inner loop to next fix mode."
			break
		fi
	done
	# End Timeout loop
	# Leaving innertries at just 1 loop since 'udevadm settle' is being used
	let innertries=1
	dropThrough=N
	if [ "$fixmode" != "done" ]
	then
		if [ "$fixmode" = "id-enabled" ]
		then
			log_msg "run id list mode fixup"
			if [ "$argids" = "Y" ]
			then
				if [ "$targetedMpathd" = "Y" ]
				then
					log_msg "*****Do cleanup_by_lun."
					cleanup_by_lun
					rc=$?
					if [ "$rc" != "0" ]
					then
						# no cleanup done
						dropThrough=Y
					fi
				else
					sancheck "-f --ids $*" $fixlogfile
				fi
			else
				sancheck "-fn --ids${listnotfound}" $fixlogfile
			fi
			fixmode=full
		else
			if [ "$fixmode" = "full" ]
			then
				log_msg "run full mode fixup"
				sancheck "-f" $fixlogfile
				rc=$?
				fixmode="none"
			else
				log_msg "Still could not find all the volumes! No more attmepts."
				fixmode="done"
			fi
		fi
	fi
done
# End outer fix loop

if [ "$numfound" != "$numvols" ]
then
	# Error case.
	log_err "The host script was unable to locate one or more FC volumes after searching for the prescribed time and doing device linkage cleanup. Number found was $numfound."
	# check for timeouts
	timeouts=$(echo -e $error_codes | grep -c "137")
	if [ "$timeouts" != "0" ]
	then
		log_err "There where $timeouts timeouts while running commands. Check the health of the FC HBA card, that it is seated properly, and the firmware levels are current."
	fi
	log_msg "All volume(s) NOT found!"
	if [ "$argids" = "N" ]
	then
		# return what we found anyway
		cat $pvmapfile
		cat $pvmapfile >> $alogfile
		sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
		cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
		rm -f $pvmapfile
		dstamp=$(date)
		log_msg "$dstamp"
		log_msg "exit 0"
		log_msg ""
		exit 0
	fi
	# debug output goes to log file because there is an issue with the amount of data that
	# the subagent can return for the CR case.
	if [ -f $pvmapfile ]
	then
		echo "Dumping $pvmapfile ..." >> $alogfile
		cat $pvmapfile >> $alogfile
		rm -f $pvmapfile
	fi
	log_msg "-------------------------------------------------------"
	log_msg "Dumping sancheck table..."
	sancheck "--ids $*" $alogfile
	log_msg "-------------------------------------------------------"
	log_msg "---------------------FAILED----------------------------"
	dstamp=$(date)
	log_msg "$dstamp"
	log_msg "exit 4"
	exit 4
else
	# Success case
	if [ "$numvols" = "0" ]
	then
		log_print "No LUNs attached to find."
	else
		log_print "All volume(s) found!"
		log_msg "wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health  device-date  using-vs"
	fi
	cat $pvmapfile
	cat $pvmapfile >> $alogfile
	sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
	cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
	rm -f $pvmapfile
	dstamp=$(date)
	log_msg "$dstamp"
	log_msg "exit 0"
	log_msg ""
	exit 0
fi
# unreachable
