#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2011, 2013 All Rights Reserved
#*
#*
# version: 2.5   [3/29/2013]
# author: csky
# sancheck.sh [-l | -f | -fn | -d ] [--ids <WWID>:<LUN>~<optional-props...> ...]
#    options: See the help text echoed out below.
#
# !Script only supported on RHEL 6!
#
# The devices that this script operates on are Fibre Channel SCSI devices.
# This version does not perform cleanup on virtual multipath devices.
#
# OPTIONS:
#    -l    verbose (long) messages
#    -f    fix mode
#    -fn   fix mode without rescanning FC bus at the end.
#    -d    delete mode (removal of specific list of LUNs)
#    --ids followed by list of volume identifying information (for checking for delete)
#
# NOTE: the "STATUS" column on the short listing is a helpful tool, but does
#       not represet mutually exclusive volume states. Some of the status
#       categories chosen may overlap and one gets chosen. If the -f (fix)
#       option is specified, the STATUS is the state prior to any fix being done.
#       Here is a brief guide:
#   STATUS         DESCRIPTION
#   ------         -----------------------------------------------------------------
#   GOOD           Block-dev is readable and a WWID path is present in
#                  /etc/disk/by-id (stable path) for block-dev.
#   DUPED          Block-dev is readable, NO stable path, but there is a stable
#                  path to the same volume (different block-dev) that is readable.
#   PASSIVE        Block-dev is NOTE readable, but there is another block dev with
#                  a matching WWID that is readable.
#   PASSIVE-D      Block-dev is NOT readable, NO stable path, but there is a stable
#                  path to the same volume that is readable. (passived-duped)
#   PASSIVE-E      Block-dev is NOT readable, NO stable path, but there is a stable
#                  path to the same volume that is NOT readable. (passive-error)
#   PASSIVE-M      Block-dev is NOT readable, NO stable path, and no alternate
#                  stable paths to same volume. (passive-missing)
#   MISSING        Block-dev is readable, NO stable path, and no alternate stable
#                  paths to the same volume.
#   MISSING(rc)    Block-dev is readable, NO stable path, there is a stable path to
#                  the same volume that is NOT readable (rc=read return code).
#   BAD(-1)        The block-dev path does NOT even exist.
#   BAD(-2)        NO page 0x83 scsi-id could be retrieved for the block-dev.
#   BAD(-3)        Block-dev may or may not be readable, there is a stable path,
#                  but NO stable paths end in the WWID obtained.
#   BAD(rc)        Block-dev is not readable (rc=read error). Stale device.
#   OFFLINE        Block-dev is readable, but is empty (no data returned). Size
#                  is effectively '0'.
#   PREMATURE      Special bad state where there is a by-id 'stable path' but it
#                  begins with scsi-S<vendor> - which is not really stable, and not
#                  the WWID udev rule we require.
#
# Fixable states: BAD(x), PREMATURE, some 'partially bad' states, non-correlating symlinks.
#
# RETURN VALUE: the number of "fixed" entries
long=false
# Log file for posting run summaries
vmcdir="/var/opt/ibm/vmc"
logfile="$vmcdir/outsancheck.log"
fix=N
argids=N
del=N
err_codes=""
cmdwait=15

#subroutines
log_msg(){
	local msg=$1
	echo "[$$] $msg" >> $logfile
	echo "[$$] $msg"
}

log_err(){
	local msg=$1
	dd=$(date)
	echo "[$$]$dd: $msg" >> $logfile
	echo "[$$]$dd: $msg" 1>&2
}

pout(){
	local msg=$1
	if [ "$long" = "true" ]
	then
		echo "[$$] $msg"
	fi
}

get_scsiid(){
	pout "timeout -s SIGKILL $cmdwait scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1"
	scsiid=$(timeout -s SIGKILL $cmdwait scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1)
	local sret=$?
	if [ "$sret" = "0" ]
	then
		scsiid=${scsiid##[ ]*}
		return 0
	else
		scsiid="-"
		err_codes="$err_codes s${sret}\n"
		pout "[scsi_id return for device=$1  **Command failed. [rc=$sret]"
		return $sret
	fi
}

dd_read(){
    if [ "$argids" = "Y" -a "$inputmatch" = "N" ]
	then
		devcheck="read check skipped"
		return 0
	fi
	dstamp=$(date)
	pout "  [$dstamp] timeout -s SIGKILL $cmdwait dd if=$1 of=/dev/null bs=32 count=1 2>&1"
	devcheck=$(timeout -s SIGKILL $cmdwait dd if=$1 of=/dev/null bs=32 count=1 2>&1)
	local sret=$?
	dstamp=$(date)
	if [ "$sret" != "0" ]
	then
		err_codes="$err_codes d${sret}\n"
	else
		devcheck=$(echo "$devcheck" | head -n 1)
	fi
	pout "  [$dstamp] RET dd read of $1 [rc=$sret]   output=$devcheck"
	return $sret
}

interp_vol(){
	pout "timeout -s SIGKILL $cmdwait file -b -s $1"
	devcheck=$(timeout -s SIGKILL 20 file -b -s $1)
	local sret=$?
	if [ "$sret" != "0" ]
	then
		err_codes="$err_codes f${sret}\n"
	fi
	pout "file -b -s $1     [rc=$sret]    output=$devcheck"
	return $sret
}

parse_vol_input(){
	# Checking for new I-T-L parms in preparation for them being passed.
	# The new params are not used by this script currently.
	local curvol=$1
	argwwid=${curvol%%:*}
	local v1a=${curvol##*:}
	arglun=${v1a%%~*}
	argblocks="-1"
	argname="[unknown-address]"
	arghba="-1"
	argtarget="-1"
	if [ "$arglun" = "$argwwid" ]
	then
		arglun="-1"
		return 0
	fi
	local v1b=${v1a#*~}
	if [ "$v1b" = "$v1a" ]
	then
		return 0
	fi
	argname=${v1b%%~*}
	local v1c=${v1b#*~}
	if [ "$v1c" = "$v1b" ]
	then
		return 0
	fi
	argblocks=${v1c%%~*}
	local v1d=${v1c#*~}
	if [ "$v1d" = "$v1c" ]
	then
		return 0
	fi
	arghba=${v1d%%~*}
	local v1e=${v1d#*~}
	if [ "$v1e" = "$v1d" ]
	then
		return 0
	fi
	argtarget=${v1d%%~*}
	return 0
}


if [ "$1" = "-l" ]
then
	long=true
fi
if [ "$1" = "-f" -o "$1" = "-fn" ]
then
	long=true
	fix=Y
fi
if [ "$1" = "-fn" ]
then
	limitrescan=Y
fi
if [ "$1" = "-d" ]
then
	del=Y
	long=true
fi
if [ "$1" = "--ids" ]
then
	shift 1
	argids=Y
else
	if [ "$2" = "--ids" ]
	then
		shift 2
		argids=Y
	fi
fi
if [ -d $vmcdir ]
then
	pout "$vmcdir exists."
else
	mkdir -p $vmcdir
	rc=$?
	pout "Created working output directory $vmcdir. rc=$rc"
fi
header=$(printf '%10s %7s %6s %10s %8s %34s %12s  %s' "SCSI-QUAD" "VENDOR" "MODEL" "BLK-DEV" "GB" "WWID" "STATE" "READ-CHECK")
if [ "$long" = "true" ]
then
	if [ -f $logfile ]
	then
		ccrc=$(find $logfile -size +5000)
		if [ "$ccrc" = "$logfile" ]
		then
			echo "Wrap the old log file."
			mv -f $logfile "${logfile}.1"
		fi
	fi
	date >> $logfile
	date
	log_msg "fix=$fix  del=$del  argids=$argids"
	pout "$@"
	log_msg "$header"
	echo "**********************************************"
	echo "  KVM SAN SCSI volume consistency checking... "
	echo "**********************************************"
	echo
else
	log_msg "fix=$fix  del=$del  argids=$argids" 
	log_msg "$header"
fi
# Setup vars
let numdisks=0
let numbad=0
let numgood=0
let numremoved=0
let numadded=0
let nummissing=0
let numduped=0
let numpassive=0
let numoffline=0
let byidremoved=0
hostBAs=""

# Initially, make a table of all scsi devices to WWID mappings.
# The table is stored in a temp file
devmapfile=$(mktemp $vmcdir/devmap.XXXXXXXX)
for sysblk in $(ls -1 -d /sys/block/sd*)
do
	devname=${sysblk/sys\/block/dev}
	get_scsiid $devname
	ret=$?
	if [ -n "$scsiid" -a  "$ret" = "0" ]
	then
		printf '%-13s %34s\n' "$devname" "$scsiid" >> $devmapfile
	fi
done
if [ "$long" = "true" ]
then
	mapped=$(cat $devmapfile | wc -l)
	dstamp=$(date)
	pout " [$dstamp]  Created a wwid mapping file with $mapped devices: $devmapfile"
	cat $devmapfile
fi

# Check for FC trasport
if [ -d /sys/class/fc_transport ]
then
	pout "Detection of FC transport folder."
else
	pout "FC transport not detected - no volumes attached to check."
fi

# loop over transport combinations (HBAs, channels, controllers)
for fcdir in $(ls -1 /sys/class/fc_transport)
do
	pout "FC transport found: $fcdir"
	target=$(echo $fcdir | cut -d : -f 1)
	channel=$(echo $fcdir | cut -d : -f 2)
	controller=$(echo $fcdir | cut -d : -f 3)
	hostnum=${target#target}
	host=host${target#target}
	# strip off previous ,host if same name
	hostBAs=${hostBAs%%,$host}
	hostBAs=${hostBAs%%$host}
	if [ -z "$hostBAs" ]
	then
		hostBAs=$host
	else
		hostBAs="$hostBAs,$host"
	fi
	portid=$(cat /sys/class/fc_transport/$fcdir/port_id)
	sportwwn=$(cat /sys/class/fc_transport/$fcdir/node_name)
	hportid=$(cat /sys/class/fc_host/$host/port_id)
	hportwwn=$(cat /sys/class/fc_host/$host/port_name)
	hportstate=$(cat /sys/class/fc_host/$host/port_state)
	if [ "$long" = "true" ]
	then
		pout "  target=$target  channel=$channel  controller=$controller  host=$host  hostnum=$hostnum"
		pout "  SAN_WWNN:   $sportwwn    PortID:  $portid"
		pout "  Host_WWNN:  $hportwwn    PortID:  $hportid    State:  $hportstate"
	fi
	
    for file in $(ls -1 /sys/class/fc_transport/$fcdir/device | egrep "^$hostnum:$channel:$controller:")
    do
		dlink=/sys/class/fc_transport/$fcdir/device/$file
		type=$(cat $dlink/type)
		state=$(cat $dlink/state)
		lun=$(echo $file | cut -d : -f 4)
		vendor=$(cat $dlink/vendor)
		vendor=${vendor%%[ ]*}
		model=$(cat $dlink/model)
		model=${model%[ ]*}
		blockdir=""
		devcheck="n/a"
		blockdev=""
		size=""
		sizegb=""
		ronly=""
		devmajor=""
		devminor=""
		capability=""
		devblockpath=""
		devpathstate=""
		dstate=""
		if [ -d $dlink/block ]
		then
			blockdev=$(ls $dlink/block)
			blockdir=$dlink/block/$blockdev
			size=$(cat $blockdir/size)
			sizegb=$(echo "scale=3; $size * 512 / ( 1024 * 1024 * 1000 )" | bc)
			ronly=$(cat $blockdir/ro)
			devmajor=$(cat $blockdir/dev | cut -d : -f 1)
			devminor=$(cat $blockdir/dev | cut -d : -f 2)
			capability=$(cat $blockdir/capability)
			devblockpath=/dev/$blockdev
			devpathstate="Missing"
			if [ -b $devblockpath ]
			then
				devpathstate="Exists"
			fi
			if [ "$ronly" = "0" ]
			then
				ronly=false
			else
				ronly=true
			fi
		else
			blockdev="n/a"
		fi
		if [ "$long" = "true" ]
		then
			echo
			pout "--------------------------------------------------"
			pout "    Examining FC device: $dlink"
			echo "      LUN               $lun"
			echo "      BlockDevice:      $blockdev"
			echo "      Type:             $type"
			echo "      DevState:         $state"
			echo "      Size:             $sizegb GB"
			echo "      ReadOnly:         $ronly"
			echo "      DeviceMajor:      $devmajor"
			echo "      DeviceMinor:      $devminor"
			echo "      Capability:       $capability"
			echo "      Vendor:           $vendor"
			echo "      Model:            $model"
			echo "      BlockFilePath:    $devblockpath"
			echo "      BlockSpecialFile: $devpathstate"
			echo
		fi
		if [ -z $blockdir ]
		then
			pout "    ***Warning: Non block device."
		fi
		let numdisks=$numdisks+1
		scsiid='[unknown]'
		volid='[unknown]'
		idmatch='false'
		checkrc="0"
		inputmatch=N
		if [ "$devpathstate" != "Exists" ]
		then
			checkrc="-1"
		fi
		get_scsiid $devblockpath
		ret=$?
		# remove first char (3) from scsi ID to get the backend VOL ID.
		volid=${scsiid:1}
		volid=${volid##[ ]*}
		if [ "$scsiid" = "" -o "$volid" = "" -o "$ret" != "0" ]
		then
			volid='[cannot retrieve]'
			# No WWID case. Try input match against input LUN
			if [ "$checkrc" != "-1" ]
			then
				checkrc="-2"
			fi
			if [ "$argids" = "Y" ]
			then
				for argid in $@
				do
					parse_vol_input "$argid"
					pout "    Input VOL check: wwid=$argwwid   blocks=$argblocks   Initiater/Target/Lun=$arghba/$argtarget/$arglun"
					if [ "$lun" = "$arglun" ]
					then
						inputmatch=Y
						pout "    Found input volume match based only on LUN: $lun"
						if [ "$del" = "Y" ]
						then
							# remove this device - special tag
							checkrc="DEL"
							pout "!!Mark device for deletion..."
						fi
						# Once we have input match, we don't have to keep looping
						break
					fi
				done
			fi
		else
			# Check against any input ids
			if [ "$argids" = "Y" ]
			then
				for argid in $@
				do
					parse_vol_input "$argid"
					pout "    Input VOL check: wwid=$argwwid   blocks=$argblocks   Initiater/Target/Lun=$arghba/$argtarget/$arglun"
					if [ "$arglun" = "$argwwid" ]
					then
						arglun="-1"
						argname="[unknown-address]"
					fi
					if [ "$argname" = "$arglun" ]
					then
						argname="[unknown-address]"
					fi
					# The wwid is expected in the proper alpha case
					if [ "$argwwid" = "$volid" ]
					then
						inputmatch=Y
						pout "    Found input volume match: $argid"
						if [ "$arglun" != "-1" -a "$arglun" != "$lun" ]
						then
							pout "      *Input LUN (external) is not a match!!"
						fi
						if [ "$del" = "Y" ]
						then
							# remove this device - special tag
							checkrc="DEL"
							pout "!!Mark device for deletion..."
						fi
						# Once we have input match, we don't have to keep looping
						break
					fi
				done
			fi
		fi
		pout "    Vol SCSI ID:   $scsiid"
		pout "    Trimmed VolID:  $volid"
		if [ "$devpathstate" = "Exists" ]
		then
			# The check is done even if the WWID is not there
			if [ "$fix" = "N" ]
			then
				interp_vol $devblockpath
				rc=$?
			else
				dd_read $devblockpath
				rc=$?
			fi

			if [ "$devcheck" = "empty" ]
			then
				let numoffline=$numoffline+1
				pout "   ***No data to read. Marking storage as offline."
			fi
			if [ "$checkrc" = "0" ]
			then
				checkrc=$rc
			fi
		fi
		fixthis=N
		if [ "$fix" = "Y" -a "$argids" = "N" ]
		then
			fixthis=Y
		else
			if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
			then
				fixthis=Y
			fi
		fi
		# If any 'check' failure at this point, remove the line from the mapping file
		# So that it is not mistaken for a good device later and checked.
		if [ "$checkrc" != "0" -a "$blockdev" != "n/a" ]
		then
			preln=$(cat $devmapfile | wc -l)
			sed -i "/${blockdev}[ ]/d" $devmapfile
			postln=$(cat $devmapfile | wc -l)
			pout "    sed -i /${blockdev}[ ]/d $devmapfile"
			pout "    Removed $blockdev row from $devmapfile. $preln lines to $postln."
		fi

		# correlate to by-id
		let numbyids=0
		if [ "$blockdev" != "n/a" ]
		then
			pout "    /dev/disk/byi-id correlation(s):"
			badmatch="N"
			for byident in $(ls -l /dev/disk/by-id | egrep "/$blockdev\$" | awk '{print $9}')
			do
				# Does byident end with volid?
				idchk=${byident%$volid}
				if [ "$idchk" != "$byident" ]
				then
				    # Since not equal, the volid portion was removed and symlink matches
					idmatch='true'
					let numbyids=$numbyids+1
					prematcheck=${byident#scsi-S$vendor}
					if [ "$prematcheck" != "$byident" ]
					then
						let numpremat=$numpremat+1
						dstate="PREMATURE"
						pout "    PREMATURE stable path link name: $byident"
					fi
				else
					if [ "$fixthis" = "Y" ]
					then
						# cleanup by-id link & rescan
						rm -f /dev/disk/by-id/$byident
						rmret=$?
						pout "  **Cleanup bad link: /dev/disk/by-id/$byident    RC=$rmret"
					fi
					# Only if device is unreadable, mark as a bad match.
					if [ "$checkrc" = "1" ]
					then
						badmatch="Y"
					fi
				fi
				pout "          $byident     (ID-Match = $idmatch)"
			done
			if [ "$numbyids" = "0" -a "$badmatch" = "Y" ]
			then
				# If the scsi block dev already exists, rescanning won't help
				# e.g. it won't cause a new /dev/sdXX and/or a new link in 'by-id', so mark for cleanup
				pout "    Mark device for cleanup: $blockdev"
				checkrc="-3"
			fi
		fi
		pout "    CheckRC=$checkrc"
		if [ "$checkrc" = "1" ]
		then
			# check for passive link case. Need to loop through all scsi devices looking for matching ids
			for chkdev in $(grep $scsiid $devmapfile | awk '{print $1}')
			do
				if [ "$chkdev" != "$devblockpath" ]
				then
					devcheck2=$(dd if=$chkdev of=/dev/null bs=32 count=1 2>&1)
					rc=$?
					if [ "$rc" = "0" ]
					then
						if [ "$dstate" = "" ]
						then
							dstate="PASSIVE"
							checkrc="0"
						fi
						let numpassive=$numpassive+1
						pout "    Another active link found to volume: $chkdev    So $blockdev is passive. Cur state: $dstate"
						break
					fi
				fi
			done
		fi
		if [ "$checkrc" = "0" -a "$dstate" = "PREMATURE" ]
		then
			# a premature state where no other issue overrides it
			checkrc="-4"
		fi
		if [ "$checkrc" = "-1" -a "$model" = "Universal Xport" ]
		then
			pout "    ***Special LUN device. Model=$model"
			dstate="SPECIAL"
			checkrc=0
			numbyids=1
		fi

		# Final judgement
		if [ "$checkrc" = "0" ]
		then
			if [ "$numbyids" != "0" ]
			then
				if [ "$dstate" != "PASSIVE" ]
				then
					if [ "$devcheck" = "empty" ]
					then
						dstate="OFFLINE"
						# FIX: This is not a fixable condition since it indicates a backend storage issue
					else
						let numgood=$numgood+1
						if [ -z "$dstate" ]
						then
							dstate="GOOD"
						fi
						pout "    ---GOOD---"
					fi
				else
					pout "    ---PASSIVE-GOOD---"
				fi
			else
				# Another block device to the same volume might have a link in by-id though.
				byidww=$(ls -l /dev/disk/by-id | awk -v ss="wwn-0x$volid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -z "$byidww" ]
				then
					 # Allow any scsi type prefix that is a single digit, not just '3' ('scsi-.' below)
					byidww=$(ls -l /dev/disk/by-id | awk -v ss="scsi-.$volid[^-]" '$0 ~ ss {print $9, $11}')
				fi
				if [ -z "$byidww" ]
				then
					pout "    ***WWID BY-ID PATH MISSING***"
					if [ "$dstate" = "PASSIVE" ]
					then
						dstate="PASSIVE-M"
					else
						# Could be corrected if there was a link in 'by-id'
						dstate="MISSING"
						let nummissing=$nummissing+1
						if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
						then
							# make a symlink. None currently exist.
							linkname="/dev/disk/by-id/wwn-0x$volid"
							ln -s $devblockpath $linkname
							rc=$?
							if [ "$rc" = "0" ]
							then
								let numadded=$numadded+1
							fi
							pout "    Adding stable symlink: ln -s $devblockpath $linkname   RC=$rc"
						fi
					fi
				else
					otherdev=$(echo $byidww | cut -d '/' -f 3)
					dev2check=$(timeout -s SIGKILL $cmdwait file -b -s /dev/$otherdev)
					rc=$?
					if [ "$long" = "true" ]
					then
						pout "    Nothing in /dev/disk/by-id points to block dev $blockdev, but there is a link for same vol:  $byidww"
						if [ "$rc" = "0" ]
						then
							pout "    $dstate ---DUPLICATE---"
						else
							pout "    ***WWID BY-ID PATH MISSING (bad symlink) $dstate***"
							# Nothing to cleanup here since this is a different device - needs to be caught in another loop
						fi
					fi
					if [ "$rc" = "0" ]
					then
						if [ "$dstate" = "PASSIVE" ]
						then
							dstate="PASSIVE-D"
						else
							dstate="DUPED"
							let numduped=$numduped+1
						fi
					else
						if [ "$dstate" = "PASSIVE" ]
						then
							dstate="PASSIVE-E"
						else
							dstate="MISSING($rc)"
							let nummissing=$nummissing+1
							# fix possible here too but old link(s) must be removed
							if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
							then
								linkname="/dev/disk/by-id/wwn-0x$volid"
								if [ -L "$linkname" ]
								then
									rm -f $linkname
									rc=$?
									pout "      REMOVE: $linkname  (rc=$rc)"
									if [ "$rc" = "0" ]
									then
										let byidremoved=$byidremoved+1
									fi
								fi
								ln -s $devblockpath $linkname
								rc=$?
								if [ "$rc" = "0" ]
								then
									let numadded=$numadded+1
								fi
								pout "    Adding stable symlink: ln -s $devblockpath $linkname   RC=$rc"
							fi
						fi
					fi
				fi
			fi
		else
			let numbad=$numbad+1
			# check for special premature state
			if [ "$dstate" != "PREMATURE" ]
			then
				dstate="BAD($checkrc)"
			fi
			if [ "$checkrc" = "DEL" ]
			then
				dstate="DELETED*"
				dstamp=$(date)
				pout "  before delete: $dstamp"
				pout "echo 1 > /sys/class/scsi_device/$file/device/delete"
				echo 1 > /sys/class/scsi_device/$file/device/delete
				rc=$?
				dstamp=$(date)
				pout "  after delete: $dstamp"
				pout "!!Remove device: echo 1 > /sys/class/scsi_device/$file/device/delete  (rc=$rc)"
				if [ "$rc" = "0" ]
				then
					let numremoved=$numremoved+1
				fi
			fi
			pout "    ***BAD/STALE*** rc=$checkrc"
			if [ "$fixthis" = "Y" ]
			then
				#Fix: remove the 'bad' device
				pout "    Removing stale device $devblockpath (echo 1 > $dlink/delete) ..."
				echo 1 > $dlink/delete
				rc=$?
				pout "    Remove rc=$rc"
				if [ "$rc" = "0" ]
				then
					let numremoved=$numremoved+1
				fi
			else
				pout "    Fix stale links not requested. Continue..."
			fi
		fi
		# table dumper
		content=${devcheck:0:42}
		model=${model%%[ ]*}
		smodel=${model:0:7}
		if [ "$inputmatch" = "Y" ]
		then
			dstate="${dstate}*"
		fi
		# SCSI-QUAD  VENDOR  MODEL  BLOCK  GB  WWID  STATE  READ-CHECK
		outvol=$(printf '%10s %7s %8s %10s %8.3f %34s %12s  %s' "$file" "$vendor" "$smodel" "$devblockpath" "$sizegb" "$volid" "$dstate" "$content")
		log_msg "$outvol"
    done
	pout "==========================================================================="
done

if [ -z "$hostBAs" ]
then
	log_err "There does not appear to be any SAN volumes attached to the host."
	exit -1
else
	# Also manually clean up stale links in /dev/disk/by-id
	if [ "$fix" = "Y" -a "$argids" = "N" -o "$del" = "Y" ]
	then
		echo
		if [ "$numremoved" != "0" ]
		then
			echo "[$$] idling 5 seconds..."
			sleep 5
		fi
		echo "[$$] Symlink checks in by-id..."
		for symlink in $(ls -1 /dev/disk/by-id | grep "wwn-0x")
		do
			device=$(ls -l /dev/disk/by-id/$symlink | awk '{print $11}' | cut -d '/' -f 3)
			checkout=$(dd if=/dev/$device of=/dev/null bs=32 count=1 2>&1)
			rc=$?
			if [ "$rc" = "0" ]
			then
				echo "[$$]   Checked OK: /dev/disk/by-id/$symlink -> /dev/$device"
			else
				rm -f /dev/disk/by-id/$symlink
				rc2=$?
				echo "[$$]   REMOVED   : /dev/disk/by-id/$symlink -> /dev/$device  (read-rc=$rc, rm-rc=$rc2)"
				if [ "$rc2" = "0" ]
				then
					let byidremoved=$byidremoved+1
				fi
			fi
		done
		for symlink in $(ls -1 /dev/disk/by-id | grep "scsi-")
		do
			device=$(ls -l /dev/disk/by-id/$symlink | awk '{print $11}' | cut -d '/' -f 3)
			checkout=$(dd if=/dev/$device of=/dev/null bs=32 count=1 2>&1)
			rc=$?
			if [ "$rc" = "0" ]
			then
				echo "[$$]   Checked OK: /dev/disk/by-id/$symlink -> /dev/$device"
			else
				rm -f /dev/disk/by-id/$symlink
				rc2=$?
				echo "[$$]   REMOVED   : /dev/disk/by-id/$symlink -> /dev/$device  (read-rc=$rc, rm-rc=$rc2)"
				if [ "$rc2" = "0" ]
				then
					let byidremoved=$byidremoved+1
				fi
			fi
		done
		let totalremoved=$numremoved+$byidremoved
		# cleanup marker file if there
		rm -f "$vmcdir/stale.sancheck"
	else
		let totalremoved=$numremoved
		if [ $numbad -gt $numremoved ]
		then
			# marker file to let the next hostvolattach know there is more stale cleanup to perform
			echo $numbad > "$vmcdir/stale.sancheck"
			echo "[$$] Added stale.sancheck marker file."
		else
			rm -f "$vmcdir/stale.sancheck"
		fi
	fi
	if [ "$totalremoved" != "0" -a "$del" = "N" -a "$limitrescan" != "Y" ]
	then
		# Removal took place -- so rescan
		echo
		echo "[$$] Since cleanup took place, rescan...."
		sleep 2
		for rescan in $(echo $hostBAs | sed s/,/\\n/g)
		do
			echo "- - -" > /sys/class/scsi_host/$rescan/scan
			echo "   echo '- - -'  > /sys/class/scsi_host/$rescan/scan"
		done
	fi

	if [ "$long" = "true" ]
	then
		echo
		echo "[$$] Host Bus Adapter list: $hostBAs"
		echo "Number volumes: $numdisks" >> $logfile
		echo "[$$] Number volumes: $numdisks"
		echo "[$$]      Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline" >> $logfile
		echo "[$$]      Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline"
		echo "[$$]      Stale devices removed: $numremoved      Links removed: $byidremoved      Stable links added: $numadded" >> $logfile
		echo "[$$]      Stale devices removed: $numremoved      Links removed: $byidremoved      Stable links added: $numadded"
	else
		echo
		echo "Number volumes: $numdisks"
		echo "     Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline"
	fi
	if [ "$long" = "true" ]
	then
		date >> $logfile
		echo "[$$] Done."
	fi
fi
rm -f $devmapfile
let totalfixed=$totalremoved+$numadded
exit $totalfixed
