#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2013 All Rights Reserved
#*
# update-multipath-conf.sh
#   version: 2.5 [05/01/2013]
#   author:  yja
#
# !Script only supported on RHEL 6!
#
#    Description: The multipathd install process creates black list exceptions if there are SAN
#                 volumes "pre-mapped" to the host. Those can be left alone. The behavior of this
#                 script is as follows:
#
#                 1. If this line is found in the top level of the blackbox stanza, then do the processing,
#                    otherwise no-op: <        wwid "*">
#                 2. make a copy of the original multipath.conf file (renamed with .save suffix).
#                 3. If there is no blacklist_exceptions section, then add one:
#                    blacklist_exceptions {
#                    }
#                 4. Add these two lines to the beginning of the blacklist_exceptions
#                    section so that it looks like this:
#                    blacklist_exceptions {
#                       wwid "36005076........................."
#                       wwid "3600a0b8........................."
#                       # pre-existing lines here
#                    }
#
#                    where
#
#                      wwid "36005076........................."
#                      allows volumes from IBM vendor family storage devices such as Storwize V7000, SVC, and XIV.
#
#                      wwid "3600a0b8........................."
#                      allows volumes from DS vendor family storage such as DS3000, DS4000, and DS5000 series.
#
#    Options:     An optional file path is passed in; otherwise, the /etc/multipath.conf file will be processed
#                 by default
#

CONFIG_FILE=/etc/multipath.conf
if [ "$#" -gt 0 ]
then
  CONFIG_FILE="$1"
fi

if [ -e $CONFIG_FILE ]
then
  echo ${0}": Processing multipath config file: "${CONFIG_FILE}
else
  echo ${0}": "${CONFIG_FILE}" does not exist."
  echo ${0}": Processing complete"
  exit 0
fi

HAS_GENERIC_BLACKLIST_ENTRY=0
EXIT_STATUS=0

check_generic_blacklist_entry() {

MY_AWK=$(cat <<'EOF'
'BEGIN {
  bb_stanza = 0;
  rc = 0;
}
END {
  exit rc;
}
{
  if (rc == 1) {
    next;
  }
  
  if ($0 ~ /^[ \t]*[^ \t]+[ \t]+\{[ \t]*$/) {
    if ($1 == "blacklist") {
      bb_stanza++;
    } else if (bb_stanza > 0) {
      bb_stanza++;
    }
  }
  else if ($0 ~ /^[ \t]*\}[ \t]*$/) {
    if (bb_stanza > 0) {
      bb_stanza--;
    }
  }
  else {
    if (bb_stanza == 1) {
      if ($1 == "wwid" && $2 == "\"*\"") {
        rc = 1;
        exit rc;
      }
    }
  }
}'
EOF
)

  # echo $MY_AWK
  eval awk $MY_AWK $CONFIG_FILE
  RC=$?

  if [ "$RC" == "1" ]
  then
    HAS_GENERIC_BLACKLIST_ENTRY=1
  elif [ "$RC" -gt "1" ]
  then
    echo ${0}": Error encountered reading: "${CONFIG_FILE}
    exit -1
  fi
}


process_multipath_conf_file() {
MY_AWK=$(cat <<'EOF'
'BEGIN {
  bb_exceptions_stanza = 0;
  bb_exceptions_stanza_exists = 0;

  bb_exceptions[0] = "\"36005076.........................\"";
  bb_exceptions[1] = "\"3600a0b8.........................\"";

  num_bb_exceptions = 2;
}
END {
  if (bb_exceptions_stanza_exists == 0) {
    printf("blacklist_exceptions {\n");
    for (i = 0; i < num_bb_exceptions; i++) {
      printf("\twwid %s\n", bb_exceptions[i]);
    }
    printf("}\n");
  }
}
{
  if ($0 ~ /^[ \t]*[^ \t]+[ \t]+\{[ \t]*$/) {
    if ($1 == "blacklist_exceptions") {
      printf("%s\n", $0);
      bb_exceptions_stanza++;
      bb_exceptions_stanza_exists = 1;

      for (i = 0; i < num_bb_exceptions; i++) {
        printf("\twwid %s\n", bb_exceptions[i]);
      }
    }
    else {
      if (bb_exceptions_stanza > 0) {
        bb_exceptions_stanza++;
      }
      printf("%s\n", $0);
    }
  }
  else if ($0 ~ /^[ \t]*\}[ \t]*$/) {
    if (bb_exceptions_stanza > 0) {
      bb_exceptions_stanza--;
    }
    printf("%s\n", $0);
  }
  else {
    if (bb_exceptions_stanza == 1) {
      for (i = 0; i < num_bb_exceptions; i++) {
        if ($1 == "wwid" && $2 == bb_exceptions[i]) {
          next;
        }
      }
    }
    printf("%s\n", $0);
  }
}'
EOF
)

  echo ${0}": Saving copy of the original multipath config file: "${CONFIG_FILE}.save
  cp -p $CONFIG_FILE $CONFIG_FILE.save
  
  if [ $? != 0 ]
  then
    echo ${0}": Error creating file: "${CONFIG_FILE}.save
    EXIT_STATUS=-1
  else
    # echo $MY_AWK
    eval awk $MY_AWK $CONFIG_FILE.save > $CONFIG_FILE
    if [ $? != 0 ]
    then
      echo ${0}": Error encountered adding required exceptions."
      EXIT_STATUS=-1
    else
      echo ${0}": Added required exceptions."
    fi
  fi
}


check_generic_blacklist_entry
if [ "$HAS_GENERIC_BLACKLIST_ENTRY" == "1" ]
then
  echo ${0}": Generic blackboxing of wwid's detected."
  process_multipath_conf_file
else
  echo ${0}": Generic blackboxing of wwid's not detected. No updates performed."
fi
echo ${0}": Processing complete"
exit $EXIT_STATUS